EXPORTS
; this is not a real def file for gdiplus library
; it is only a def file with symbols strictly neede
; to make link gdk-pixbuf with gdiplus.dll on systems
; where the import library does not exit and it contains
; symbols for both win32 and win64 compilations in 
; the same def file, since win32 uses __stdcall calling
; convention and win64 uses __cdecl convention; the symbols
; are nevertheless the same
	GdipBitmapGetPixel
	GdipBitmapGetPixel@16
	GdipBitmapSetPixel
	GdipBitmapSetPixel@16
	GdipBitmapSetResolution
	GdipBitmapSetResolution@12
	GdipCreateBitmapFromScan0
	GdipCreateBitmapFromScan0@24
	GdipCreateBitmapFromStream
	GdipCreateBitmapFromStream@8
	GdipDeleteGraphics
	GdipDeleteGraphics@4
	GdipDisposeImage
	GdipDisposeImage@4
	GdipDrawImageI
	GdipDrawImageI@16
	GdipFlush
	GdipFlush@8
	GdipGetImageEncoders
	GdipGetImageEncoders@12
	GdipGetImageEncodersSize
	GdipGetImageEncodersSize@8
	GdipGetImageFlags
	GdipGetImageFlags@8
	GdipGetImageGraphicsContext
	GdipGetImageGraphicsContext@8
	GdipGetImageHeight
	GdipGetImageHeight@8
	GdipGetImageHorizontalResolution
	GdipGetImageHorizontalResolution@8
	GdipGetImageVerticalResolution
	GdipGetImageVerticalResolution@8
	GdipGetImageWidth
	GdipGetImageWidth@8
	GdipGetPropertyItem
	GdipGetPropertyItem@16
	GdipGetPropertyItemSize
	GdipGetPropertyItemSize@12
	GdipGraphicsClear
	GdipGraphicsClear@8
	GdipImageGetFrameCount
	GdipImageGetFrameCount@12
	GdipImageSelectActiveFrame
	GdipImageSelectActiveFrame@12
	GdipLoadImageFromStream
	GdipLoadImageFromStream@8
	GdiplusStartup
	GdiplusStartup@12
	GdipSaveImageToStream
	GdipSaveImageToStream@16
