#include <winver.h>
#include "rcVersion.h"

VS_VERSION_INFO		VERSIONINFO
  FILEVERSION		LIBXML_MAJOR_VERSION,LIBXML_MINOR_VERSION,LIBXML_MICRO_VERSION,0
  PRODUCTVERSION	LIBXML_MAJOR_VERSION,LIBXML_MINOR_VERSION,LIBXML_MICRO_VERSION,0
  FILEFLAGSMASK		VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS		VS_FF_DEBUG
#else
  FILEFLAGS		0
#endif
  FILEOS		VOS__WINDOWS32
  FILETYPE		VFT_DLL
  FILESUBTYPE	VFT2_UNKNOWN	// not used
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "04090000"  /* Lang = US English, Charset = ASCII */
    BEGIN
      VALUE "FileDescription",	"libxml2 library\0"
      VALUE "FileVersion",	LIBXML_DOTTED_VERSION "\0"
      VALUE "InternalName",	"libxml2.dll\0"
      VALUE "LegalCopyright",	"Copyright (C) Daniel Veillard\0"
      VALUE "LegalTrademarks",	"\0"
      VALUE "OriginalFilename",	"libxml2.dll\0"
      VALUE "ProductName",	"libxml2\0"
      VALUE "ProductVersion",	LIBXML_DOTTED_VERSION "\0"
      VALUE "Comments",		"For more information visit https://gitlab.gnome.org/GNOME/libxml2\0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
      VALUE "Translation", 0x0409, 0  /* US English, ASCII */
  END
END
